/*
 * reload1.c - Bootstrap and reset the PS2 OS from the RTE.
 *
 * Copyright (c) 2003 M. R. Brown <mrbrown@0xd6.org>
 *
 * This code is licensed under the LGPL version 2.  See the file
 * "COPYING.LGPL" in the main directory of this archive for more information.
 */

#include "types.h"
#include "defs.h"

#define ELF_MAGIC	0x464c457f
#define ELF_PT_LOAD	1

struct elf_header {
	u8	ident[16];
	u16	type;
	u16	machine;
	u32	version;
	u32	entry;
	u32	phoff;
	u32	shoff;
	u32	flags;
	u16	ehsize;
	u16	phentsize;
	u16	phnum;
	u16	shentsize;
	u16	shnum;
	u16	shstrndx;
};

struct elf_pheader {
	u32	type;
	u32	offset;
	void	*vaddr;
	u32	paddr;
	u32	filesz;
	u32	memsz;
	u32	flags;
	u32	align;
};

/* You will want to change this for a T10k */
static const u32 ramsize = 32 * 1024 * 1024;

extern void *_binary_elf_start;

/* From romfs.c */
int romfs_findfile(const char *filename, u32 *offset, u32 *size,
		u32 *ext_offset);

/* From rtesif.c */
void sif_init();
int sif_reg_set(int, int);
int sif_reg_get(int);
int sif_dma_request(void *, int);

static int eeload_setup(char *name, u32 pc, u32 sp, u32 val);
static void reset_iop();

void _start()
{
	u32 kaddr, ksize, i;

	asm ("li\t$29,%0\n\t"::"i" (0x100000));

	/* Reboot the IOP first. */
	reset_iop();

	asm("di");

	if (!romfs_findfile("KERNEL\0\0\0", &kaddr, &ksize, NULL))
		goto spin;

	/* Run uncached, to avoid flushing the caches later */
	asm (""			\
	"la\t$8,1f\n\t"		\
	"lui\t$9,0x2000\n\t"	\
	"or\t$8,$9\n\t"		\
	"jr\t$8\n\t"		\
	"nop\n\t1:\n\t":::"$8", "$9");

	/*
	 * Clear kernel RAM up to where we sit, then load the kernel itself.
	 */
	for (i = 0; i < 0x30000 ; i += 16)
		_sq(0, 0xa0000000 + i);

	for (i = 0; i < ksize; i += 16)
		_sq(_lq(kaddr + i), 0xa0000000 + i);

	/*
	 * Setup the kernel parameters.
	 */
	_sw(ramsize, 0x70003ff0);

	/*
	 * This code is BIOS dependant.  From what I've heard, it doesn't work
	 * on the 3900x series.  Note that this is only one method of trapping
	 * the kernel boot sequence, I'm looking for a more portable way to do
	 * this.
	 */
	_sw(0x0c000000 | (0x3ffffff & ((u32)eeload_setup >> 2)), 0xa00010b0);

	/*
	 * Execute the kernel startup routine.
	 */
	asm ("j\t%0" : : "i" (0xa0001000));

spin:
	while (1) ;
}

/*
 * This routine is called during the kernel boot sequence to load the system
 * bootloader "EELOAD".  In addition to loading EELOAD into RAM, the original
 * routine initializes the thread information for that block of code.
 *
 * Instead of loading EELOAD, if we have an ELF file attached we call the
 * thread setup routine for it's entry point, and then proceed to load the
 * ELF into it's final RAM address.
 */
static int eeload_setup(char *filename, u32 pc, u32 sp, u32 val)
{
	int (*_thread_setup)(char *, u32, u32, u32);
	void (*_cache_func)();
	struct elf_header *eh = (struct elf_header *)&_binary_elf_start;
	struct elf_pheader *eph;
	void *source;
	int index, i;

	/* possibly BIOS dependent */
	(int (**)(char *, u32, u32, u32))_thread_setup = 0x80005390;

	if (_lw((u32)&eh->ident) != ELF_MAGIC)
		return _thread_setup(filename, pc, sp, val);

	/* 
	 * Spend as little time as possible here, load the dummy 16-byte
	 * ROMVER instead of EELOAD.
	 */
	index = _thread_setup("ROMVER", eh->entry, ramsize - 0x10000, val);

	eph = (struct elf_pheader *)((u32)&_binary_elf_start + eh->phoff);
	/*
	 * Scan through the ELF's program headers and copy them into RAM, then
	 * zero out any non-loaded regions.
	 */
	for (i = 0; i < eh->phnum; ++i) {
		if (eph[i].type != ELF_PT_LOAD)
			continue;
		source = (void *)((u32)&_binary_elf_start + eph[i].offset);
		slow_memcpy(eph[i].vaddr, source, eph[i].filesz);
		if (eph[i].memsz > eph[i].filesz)
			slow_memset(eph[i].vaddr + eph[i].filesz, 0,
					eph[i].memsz - eph[i].filesz);
	}

	/* Writeback the data cache */
	(void (**)())_cache_func = 0x80002ac0;
	_cache_func();
	/* Invalidate the insn cache */
	(void (**)())_cache_func = 0x80002a80;
	_cache_func();

	return index;
}

/*
 * A vastly simplified IOP reset routine.
 */
static void reset_iop()
{
	u8 packet[104] __attribute__((aligned(16)));
	struct dma_request {
		void *src;
		void *dest;
		u32 size;
		u32 mode;
	} dmareq;
	u8 *upkt = KSEG1ADDR((u8 *)&packet);
	int i;

	sif_init();

	for (i = 0; i < sizeof(packet); i += 8)
		_sd(0, (u32)(upkt + i));

	/* Set the size and command ID */
	_sw(sizeof(packet), (u32)upkt);
	_sw(0x80000003, (u32)(upkt + 8));

	dmareq.src  = upkt;
	dmareq.dest = (void *)sif_reg_get(2);
	dmareq.size = sizeof(packet);
	dmareq.mode = 0x44;

	if (!sif_dma_request(&dmareq, 1)) {
		while (1) ;
	}

	sif_reg_set(4, 0x10000);
	sif_reg_set(4, 0x20000);

	/* Wait for the IOP to come back */
	while (!(sif_reg_get(4) & 0x40000))
		;
	sif_reg_set(4, 0x40000);
}
